"""Command line interface definition."""

import logging
from pathlib import Path
from typing import IO, Any, List, Optional, Sequence, Tuple, Union

import click

# Migrate away from xdg to xdg-base-dirs once only Python >= 3.10 is supported
# https://github.com/lyz-code/autoimport/issues/239
import xdg
from maison.config import ProjectConfig

from autoimport import services, version

log = logging.getLogger(__name__)


def get_files(source_path: str) -> List[IO[Any]]:
    """Get all files recursively from the given source path."""
    files = []
    for py_file in Path(source_path).glob("**/*.py"):
        if py_file.is_file():
            files.append(click.File("r+").convert(py_file, None, None))
    return files


def flatten(seq: Sequence[Any]) -> Tuple[Any, ...]:
    """Flatten nested sequences."""
    flattened = []
    for items in seq:
        if isinstance(items, (tuple, list)):
            for item in items:
                flattened.append(item)
        else:
            item = items
            flattened.append(item)
    return tuple(flattened)


class FileOrDir(click.ParamType):
    """Custom parameter type that accepts either a directory or file."""

    def convert(
        self,
        value: Union[str, Path],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> List[IO[Any]]:
        """Convert the value to the correct type."""
        try:
            return [click.File("r+").convert(value, param, ctx)]
        except click.BadParameter:
            path = click.Path(exists=True).convert(value, param, ctx)
            return get_files(str(path))


@click.command()
@click.version_option(version="", message=version.version_info())
@click.option("--config-file", default=None)
@click.option("--ignore-init-modules", is_flag=True, help="Ignore __init__.py files.")
@click.option(
    "--keep-unused-imports",
    is_flag=True,
    help="If True, retains unused imports.",
    default=False,
)
@click.argument("files", type=FileOrDir(), nargs=-1)
def cli(
    files: List[IO[Any]],
    config_file: Optional[str] = None,
    ignore_init_modules: bool = False,
    keep_unused_imports: bool = False,
) -> None:
    """Corrects the source code of the specified files."""
    # Compose configuration
    config_files: List[str] = []

    global_config_path = xdg.xdg_config_home() / "autoimport" / "config.toml"
    if global_config_path.is_file():
        config_files.append(str(global_config_path))

    config_files.append("pyproject.toml")

    if config_file is not None:
        config_files.append(config_file)

    config = ProjectConfig(
        project_name="autoimport", source_files=config_files, merge_configs=True
    ).to_dict()

    # Process inputs
    flattened_files = flatten(files)
    if ignore_init_modules:
        flattened_files = tuple(
            file for file in flattened_files if "__init__.py" not in file.name
        )

    try:
        fixed_code = services.fix_files(flattened_files, config, keep_unused_imports)
    except FileNotFoundError as error:
        log.error(error)

    if fixed_code is not None:
        print(fixed_code, end="")


if __name__ == "__main__":  # pragma: no cover
    cli()  # pylint: disable=E1120



"""Entrypoint to allow running as `python3 -m autoimport`."""

from autoimport.entrypoints.cli import cli

if __name__ == "__main__":
    # this needs no-value-for-parameter exclusion because pylint
    # isn't smart enough to realize that click decorator handles it (but mypy is)
    cli()  # pylint: disable=no-value-for-parameter



"""Utilities to retrieve the information of the program version."""

import platform
import sys
from textwrap import dedent

# Do not edit the version manually, let `make bump` do it.

__version__ = "1.6.0"


def version_info() -> str:
    """Display the version of the program, python and the platform."""
    return dedent(
        f"""\
        ------------------------------------------------------------------
             autoimport: {__version__}
             Python: {sys.version.split(" ", maxsplit=1)[0]}
             Platform: {platform.platform()}
        ------------------------------------------------------------------"""
    )



"""Test the command line interface."""

import os
import re
from pathlib import Path
from textwrap import dedent
from typing import Dict, List, Optional

import pytest
from click.testing import CliRunner

from autoimport.entrypoints.cli import cli
from autoimport.version import __version__


@pytest.fixture(name="runner")
def fixture_runner() -> CliRunner:
    """Configure the Click cli test runner."""
    return CliRunner(mix_stderr=False, env={"XDG_CONFIG_HOME": "/dev/null"})


def test_version(runner: CliRunner) -> None:
    """Prints program version when called with --version."""
    result = runner.invoke(cli, ["--version"])

    assert result.exit_code == 0
    assert re.search(
        rf" *autoimport: {__version__}\n *Python: .*\n *Platform: .*",
        result.stdout,
    )


def test_corrects_one_file(runner: CliRunner, tmp_path: Path) -> None:
    """Correct the source code of a file."""
    test_file = tmp_path / "source.py"
    test_file.write_text("os.getcwd()")
    fixed_source = dedent(
        """\
        import os


        os.getcwd()"""
    )

    result = runner.invoke(cli, [str(test_file)])

    assert result.exit_code == 0
    assert test_file.read_text() == fixed_source


@pytest.mark.secondary()
def test_corrects_three_files(runner: CliRunner, tmp_path: Path) -> None:
    """Correct the source code of multiple files."""
    test_files = []
    for file_number in range(3):
        test_file = tmp_path / f"source_{file_number}.py"
        test_file.write_text("os.getcwd()")
        test_files.append(test_file)
    fixed_source = dedent(
        """\
        import os


        os.getcwd()"""
    )

    result = runner.invoke(cli, [str(test_file) for test_file in test_files])

    assert result.exit_code == 0
    for test_file in test_files:
        assert test_file.read_text() == fixed_source


def test_correct_all_files_in_dir_recursively(
    runner: CliRunner, test_dir: Path
) -> None:
    """Ensure files and dirs can be parsed and fixes associated files."""
    result = runner.invoke(cli, [str(test_dir)])

    assert result.exit_code == 0
    fixed_source = "import os\n\n\nos.getcwd()"
    assert (test_dir / "test_file1.py").read_text() == fixed_source
    assert (test_dir / "subdir/test_file2.py").read_text() == fixed_source


def test_correct_mix_dir_and_files(
    runner: CliRunner, test_dir: Path, tmp_path: Path
) -> None:
    """Ensure all files in a given directory get fixed by autoimport."""
    test_file = tmp_path / "source.py"
    test_file.write_text("os.getcwd()")

    result = runner.invoke(cli, [str(test_dir), str(test_file)])

    assert result.exit_code == 0
    fixed_source = "import os\n\n\nos.getcwd()"
    assert (test_dir / "test_file1.py").read_text() == fixed_source
    assert (test_dir / "subdir/test_file2.py").read_text() == fixed_source
    assert test_file.read_text() == fixed_source


def test_corrects_code_from_stdin(runner: CliRunner) -> None:
    """Correct the source code passed as stdin."""
    source = "os.getcwd()"
    fixed_source = dedent(
        """\
        import os


        os.getcwd()"""
    )

    result = runner.invoke(cli, ["-"], input=source)

    assert result.exit_code == 0
    assert result.stdout == fixed_source


def test_pyproject_common_statements(runner: CliRunner, tmp_path: Path) -> None:
    """Allow common_statements to be defined in pyproject.toml"""
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        dedent(
            """\
            [tool.autoimport]
            common_statements = { "FooBar" = "from baz.qux import FooBar" }
            """
        )
    )
    test_file = tmp_path / "source.py"
    test_file.write_text("FooBar\n")
    # AAA03: Until https://github.com/jamescooke/flake8-aaa/issues/196 is fixed
    with runner.isolated_filesystem(temp_dir=tmp_path):
        result = runner.invoke(cli, [str(test_file)])  # noqa: AAA03

    assert result.exit_code == 0
    assert test_file.read_text() == dedent(
        """\
        from baz.qux import FooBar


        FooBar
        """
    )


@pytest.mark.skip("Until https://github.com/dbatten5/maison/issues/141 is fixed")
def test_config_path_argument(runner: CliRunner, tmp_path: Path) -> None:
    """Allow common_statements to be defined in pyproject.toml"""
    config_dir = tmp_path / "config"
    config_dir.mkdir()
    pyproject_toml = config_dir / "pyproject.toml"
    pyproject_toml.write_text(
        dedent(
            """\
            [tool.autoimport]
            common_statements = { "FooBar" = "from baz.qux import FooBar" }
            """
        )
    )
    code_dir = tmp_path / "code"
    code_dir.mkdir()
    test_file = code_dir / "source.py"
    test_file.write_text("FooBar\n")

    result = runner.invoke(cli, ["--config-file", str(pyproject_toml), str(test_file)])

    assert result.exit_code == 0
    assert test_file.read_text() == dedent(
        """\
        from baz.qux import FooBar


        FooBar
        """
    )


@pytest.mark.parametrize(
    ("create_global_conf", "use_local_conf", "create_pyproject", "expected_imports"),
    [
        pytest.param(True, False, False, "from g import G", id="global"),
        pytest.param(False, True, False, "from r import R", id="local"),
        pytest.param(False, False, True, "from p import P", id="pyproject"),
        pytest.param(
            True, True, False, "from g import G\nfrom r import R", id="global-and-local"
        ),
        pytest.param(
            True,
            False,
            True,
            "from g import G\nfrom p import P",
            id="global-and-pyproject",
        ),
        pytest.param(
            False,
            True,
            True,
            "from r import R\nfrom p import P",
            id="local-and-pyproject",
        ),
        pytest.param(
            True,
            True,
            True,
            "from g import G\nfrom r import R\nfrom p import P",
            id="global-and-local-and-pyproject",
        ),  # noqa: R0913, R0914
    ],
)
# R0913: Too many arguments (6/5): We need to refactor this test in many more
# R0914: Too many local variables (16/15): We need to refactor this test in many more
def test_global_and_local_config(  # noqa: R0913, R0914
    runner: CliRunner,
    tmp_path: Path,
    create_global_conf: bool,
    use_local_conf: bool,
    create_pyproject: bool,
    expected_imports: str,
) -> None:
    """
    Test interaction between the following:
      - presence of the global config file $XDG_CONFIG_HOME/autoimport/config.toml
      - use of the --config-file flag to specify a local config file
      - presence of a pyproject.toml file
    """
    config = {
        "global": '[common_statements]\n"G" = "from g import G"',
        "local": '[common_statements]\n"R" = "from r import R"',
        "pyproject": '[tool.autoimport.common_statements]\n"P" = "from p import P"',
    }
    code_path = tmp_path / "code.py"
    original_code = dedent(
        """
        G
        R
        P
        """
    )
    code_path.write_text(original_code)
    args: List[str] = [str(code_path)]
    env: Dict[str, Optional[str]] = {}
    if create_global_conf:
        xdg_home = (tmp_path / "xdg_home").resolve()  # must be absolute path
        env["XDG_CONFIG_HOME"] = str(xdg_home)
        global_conf_path = xdg_home / "autoimport" / "config.toml"
        global_conf_path.parent.mkdir(parents=True)
        global_conf_path.write_text(config["global"])
    if use_local_conf:
        local_conf_path = tmp_path / "cfg" / "local.toml"
        local_conf_path.parent.mkdir(parents=True)
        local_conf_path.write_text(config["local"])
        args.extend(["--config-file", str(local_conf_path)])
    if create_pyproject:
        pyproject_path = tmp_path / "pyproject.toml"
        pyproject_path.write_text(config["pyproject"])
    # AAA03: Until https://github.com/jamescooke/flake8-aaa/issues/196 is fixed
    with runner.isolated_filesystem(temp_dir=tmp_path):
        result = runner.invoke(cli, args, env=env)  # noqa: AAA03

    assert result.exit_code == 0
    assert code_path.read_text() == expected_imports + "\n\n" + original_code


def test_global_and_local_config_precedence(runner: CliRunner, tmp_path: Path) -> None:
    """
    Test precedence of configuration specified in the global config vs
    pyproject.toml vs --config-file. From low to high priority:
      - global config file
      - project-local pyproject.toml file
      - file specified by the --config-file flag, if any
    """
    config = {
        "global": dedent(
            """
            [common_statements]
            "G" = "from g import G"
            "A" = "from ga import A"
            "B" = "from gb import B"
            "C" = "from gc import C"
            """
        ),
        "pyproject": dedent(
            """
            [tool.autoimport.common_statements]
            "A" = "from pa import A"
            "C" = "from pc import C"
            "D" = "from pd import D"
            """
        ),
        "local": dedent(
            """
            [common_statements]
            "B" = "from lb import B"
            "C" = "from lc import C"
            "D" = "from ld import D"
            """
        ),
    }
    code_path = tmp_path / "code.py"
    original_code = dedent(
        """
        A
        B
        C
        D
        G
        """
    )
    expected_imports = dedent(
        """\
        from pa import A
        from lb import B
        from lc import C
        from ld import D
        from g import G
        """
    )
    code_path.write_text(original_code)
    args: List[str] = [str(code_path)]
    env: Dict[str, Optional[str]] = {}
    # create_global_conf:
    xdg_home = (tmp_path / "xdg_home").resolve()  # must be absolute path
    env["XDG_CONFIG_HOME"] = str(xdg_home)
    global_conf_path = xdg_home / "autoimport" / "config.toml"
    global_conf_path.parent.mkdir(parents=True)
    global_conf_path.write_text(config["global"])
    # create:
    local_conf_path = tmp_path / "cfg" / "local.toml"
    local_conf_path.parent.mkdir(parents=True)
    local_conf_path.write_text(config["local"])
    args.extend(["--config-file", str(local_conf_path)])
    # create_pyproject:
    pyproject_path = tmp_path / "pyproject.toml"
    pyproject_path.write_text(config["pyproject"])
    # AAA03: Until https://github.com/jamescooke/flake8-aaa/issues/196 is fixed
    with runner.isolated_filesystem(temp_dir=tmp_path):
        result = runner.invoke(cli, args, env=env)  # noqa: AAA03

    assert result.exit_code == 0
    assert code_path.read_text() == expected_imports + "\n" + original_code


def test_fix_files_doesnt_touch_the_file_if_its_not_going_to_change_it(
    runner: CliRunner, tmp_path: Path
) -> None:
    """
    Given: A file that doesn't need any change
    When: fix files is run
    Then: The file is untouched
    """
    test_file = tmp_path / "source.py"
    test_file.write_text("a = 1")
    modified_time = os.path.getmtime(test_file)

    result = runner.invoke(cli, [str(test_file)])

    assert result.exit_code == 0
    assert os.path.getmtime(test_file) == modified_time



"""Tests the service layer."""

from textwrap import dedent

import pytest

from autoimport.model import common_statements
from autoimport.services import fix_code


def test_fix_code_adds_missing_import() -> None:
    """Understands that os is a package and add it to the top of the file."""
    source = "os.getcwd()"
    fixed_source = dedent(
        """\
        import os


        os.getcwd()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_doesnt_change_source_if_package_doesnt_exist() -> None:
    """As foo is not found, nothing is changed."""
    source = "foo"

    result = fix_code(source)

    assert result == source


def test_fix_imports_packages_below_docstring() -> None:
    """Imports are located below the module docstrings."""
    source = dedent(
        '''\
        """Module docstring.

        """
        import pytest
        os.getcwd()'''
    )
    fixed_source = dedent(
        '''\
        """Module docstring.

        """

        import os


        os.getcwd()'''
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_imports_packages_below_single_line_docstring() -> None:
    """Imports are located below the module docstrings when they only take one line."""
    source = dedent(
        '''\
        """Module docstring."""

        import pytest
        os.getcwd()'''
    )
    fixed_source = dedent(
        '''\
        """Module docstring."""

        import os


        os.getcwd()'''
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_imports_type_hints() -> None:
    """Typing objects are initialized with their required header."""
    source = dedent(
        """\
        def function(dictionary: Dict) -> None:
            pass"""
    )
    fixed_source = dedent(
        """\
        from typing import Dict


        def function(dictionary: Dict) -> None:
            pass"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_removes_unneeded_imports() -> None:
    """If there is an import statement of an unused package it should be removed."""
    source = dedent(
        """\
        import requests
        foo = 1"""
    )
    fixed_source = "foo = 1"

    result = fix_code(source)

    assert result == fixed_source


def test_fix_removes_multiple_unneeded_imports() -> None:
    """
    Given: A source code with multiple unused import statements.
    When: fix_code is run.
    Then: The unused import statements are deleted.
    """
    source = dedent(
        """\
        import requests
        from textwrap import dedent

        from yaml import YAMLError
        foo = 1"""
    )
    fixed_source = "foo = 1"

    result = fix_code(source)

    assert result == fixed_source


def test_fix_removes_unneeded_imports_in_from_statements() -> None:
    """Remove `from package import` statement of an unused packages."""
    source = dedent(
        """\
        from os import path
        foo = 1"""
    )
    fixed_source = "foo = 1"

    result = fix_code(source)

    assert result == fixed_source


def test_fix_removes_unused_imports_in_multiline_from_statements() -> None:
    """
    Given: A source code with multiline import from statements.
    When: fix_code is run
    Then: Unused import statements are deleted
    """
    source = dedent(
        """\
        from os import (
            getcwd,
            path,
        )

        getcwd()"""
    )
    fixed_source = dedent(
        """\
        from os import (
            getcwd,
        )


        getcwd()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_removes_unneeded_imports_in_beginning_of_from_statements() -> None:
    """Remove unused `object_name` in `from package import object_name, other_object`
    statements.
    """
    source = dedent(
        """\
        from os import path, getcwd

        getcwd()"""
    )
    fixed_source = dedent(
        """\
        from os import getcwd


        getcwd()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_removes_unneeded_imports_in_middle_of_from_statements() -> None:
    """Remove unused `object_name` in
    `from package import other_object, object_name, other_used_object` statements.
    """
    source = dedent(
        """\
        from os import getcwd, path, mkdir

        getcwd()
        mkdir()"""
    )
    fixed_source = dedent(
        """\
        from os import getcwd, mkdir


        getcwd()
        mkdir()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_removes_unneeded_imports_in_end_of_from_statements() -> None:
    """Remove unused `object_name` in `from package import other_object, object_name`
    statements.
    """
    source = dedent(
        """\
        from os import getcwd, path

        getcwd()"""
    )
    fixed_source = dedent(
        """\
        from os import getcwd


        getcwd()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_respects_multiple_from_import_lines() -> None:
    """
    Given: Multiple from X import Y lines.
    When: Fix code is run
    Then: The import statements aren't broken
    """
    source = dedent(
        """\
        from os import getcwd

        from re import match


        getcwd()
        match(r'a', 'a')"""
    )

    result = fix_code(source)

    assert result == source


def test_fix_respects_multiple_from_import_lines_in_multiple_lines() -> None:
    """
    Given: Multiple from X import Y lines, some with multiline format.
    When: Fix code is run
    Then: The import statements aren't broken
    """
    source = dedent(
        """\
        from os import (
            getcwd,
        )

        from re import match


        getcwd()
        match(r'a', 'a')"""
    )

    result = fix_code(source)

    assert result == source


def test_fix_respects_import_lines_in_multiple_line_strings() -> None:
    """
    Given: Import lines in several multiline strings.
    When: Fix code is run.
    Then: The import statements inside the string are not moved to the top.
    """
    source = dedent(
        """\
        from textwrap import dedent

        source = dedent(
            \"\"\"\\
            from re import match

            match(r'a', 'a')\"\"\"
        )

        source = dedent(
            \"\"\"\\
            from os import (
                getcwd,
            )

            getcwd()\"\"\"
        )"""
    )
    fixed_source = dedent(
        """\
        from textwrap import dedent


        source = dedent(
            \"\"\"\\
            from re import match

            match(r'a', 'a')\"\"\"
        )

        source = dedent(
            \"\"\"\\
            from os import (
                getcwd,
            )

            getcwd()\"\"\"
        )"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_moves_import_statements_to_the_top() -> None:
    """Move import statements present in the source code to the top of the file"""
    source = dedent(
        """\
        a = 3

        import os
        os.getcwd()"""
    )
    fixed_source = dedent(
        """\
        import os


        a = 3

        os.getcwd()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_moves_import_statements_in_indented_code_to_the_top() -> None:
    """Move import statements present indented in the source code
    to the top of the file
    """
    source = dedent(
        """\
        import requests

        requests.get('hi')

        def test():
            import os
            os.getcwd()"""
    )
    fixed_source = dedent(
        """\
        import requests

        import os


        requests.get('hi')

        def test():
            os.getcwd()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_skips_moves_to_the_top_when_disabled_is_true() -> None:
    """Skip moving import statements when disable_move_to_top config is true."""
    source = dedent(
        """\
        import requests


        requests.get('hi')

        def test():
            import os
            os.getcwd()"""
    )
    config = {"disable_move_to_top": True}

    result = fix_code(source, config)

    assert result == source


def test_fix_skips_moves_to_the_top_when_disabled_is_false() -> None:
    """Moving import statements should still occur when disable_move_to_top
    config is false.
    """
    source = dedent(
        """\
        import requests

        requests.get('hi')

        def test():
            import os
            os.getcwd()"""
    )
    fixed_source = dedent(
        """\
        import requests

        import os


        requests.get('hi')

        def test():
            os.getcwd()"""
    )
    config = {"disable_move_to_top": False}

    result = fix_code(source, config)

    assert result == fixed_source


def test_fix_moves_from_import_statements_to_the_top() -> None:
    """Move from import statements present in the source code to the top of the file"""
    source = dedent(
        """\
        a = 3

        from os import getcwd
        getcwd()"""
    )
    fixed_source = dedent(
        """\
        from os import getcwd


        a = 3

        getcwd()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_moves_multiline_import_statements_to_the_top() -> None:
    """
    Given: Multiple from X import Y lines.
    When: Fix code is run.
    Then: The import statements are moved to the top.
    """
    source = dedent(
        """\
        from os import getcwd

        getcwd()

        from re import (
            match,
        )
        match(r'a', 'a')"""
    )
    fixed_source = dedent(
        """\
        from os import getcwd

        from re import (
            match,
        )


        getcwd()

        match(r'a', 'a')"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_doesnt_break_objects_with_import_in_their_names() -> None:
    """Objects that have the import name in their name should not be changed."""
    source = dedent(
        """\
        def import_code():
            pass

        def code_import():
            pass

        def import():
            pass

        import_string = 'a'"""
    )

    result = fix_code(source)

    assert result == source


def test_fix_doesnt_move_import_statements_with_noqa_to_the_top() -> None:
    """Ignore lines that have # noqa: autoimport."""
    source = dedent(
        """\
        a = 3

        from os import getcwd # noqa: autoimport
        getcwd()"""
    )

    result = fix_code(source)

    assert result == source


def test_fix_doesnt_fail_on_noqa_lines_on_unused_import() -> None:
    """Ignore lines that have # noqa: autoimport."""
    source = dedent(
        """\
        from os import getcwd # noqa: autoimport"""
    )

    result = fix_code(source)

    assert result == source


def test_fix_respects_fmt_skip_lines() -> None:
    """Ignore lines that have # fmt: skip."""
    source = dedent(
        """
        def why():
            import pdb;pdb.set_trace()  # fmt: skip
            return 'dunno'
        """
    ).replace("\n", "", 1)

    result = fix_code(source)

    assert result == source


def test_fix_respects_noqa_in_from_import_lines_in_multiple_lines() -> None:
    """
    Given: Multiple from X import Y lines, some with multiline format with noqa
        statement.
    When: Fix code is run.
    Then: The import statements aren't broken.
    """
    source = dedent(
        """\
        from os import getcwd


        getcwd()

        from re import ( # noqa: autoimport
            match,
        )

        match(r'a', 'a')"""
    )

    result = fix_code(source)

    assert result == source


def test_fix_respects_strings_with_import_statements() -> None:
    """
    Given: Code with a string that has import statements structure.
    When: Fix code is run.
    Then: The string is respected
    """
    source = dedent(
        """\
        import_string = 'import requests'
        from_import_string = "from re import match"
        multiline string = dedent(
            \"\"\"\\
            import requests
            from re import match
            \"\"\"
        )
        multiline single_quote_string = dedent(
            \'\'\'\\
            import requests
            from re import match
            \'\'\'
        )
        import os"""
    )
    fixed_source = dedent(
        """\
        import os


        import_string = 'import requests'
        from_import_string = "from re import match"
        multiline string = dedent(
            \"\"\"\\
            import requests
            from re import match
            \"\"\"
        )
        multiline single_quote_string = dedent(
            \'\'\'\\
            import requests
            from re import match
            \'\'\'
        )"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_doesnt_mistake_docstrings_with_multiline_string() -> None:
    """
    Given: A function with a docstring.
    When: Fix code is run.
    Then: The rest of the file is not mistaken for a long multiline string
    """
    source = dedent(
        """\
        def function_1():
            \"\"\"Function docstring\"\"\"
            import os
            os.getcwd()"""
    )
    fixed_source = dedent(
        """\
        import os


        def function_1():
            \"\"\"Function docstring\"\"\"
            os.getcwd()"""
    )

    result = fix_code(source)

    assert result == fixed_source


@pytest.mark.parametrize(
    ("import_key", "import_statement"),
    ((key, value) for key, value in common_statements.items()),
    ids=list(common_statements.keys()),
)
def test_fix_autoimports_common_imports(import_key: str, import_statement: str) -> None:
    """
    Given: Code with missing import statements that match the common list.
    When: Fix code is run.
    Then: The imports are done
    """
    source = dedent(
        f"""\
        import os

        os.getcwd

        variable = {import_key}"""
    )
    fixed_source = dedent(
        f"""\
        import os

        {import_statement}


        os.getcwd

        variable = {import_key}"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_autoimports_objects_defined_in_the_root_of_the_package() -> None:
    """
    Given:
        The fix code is run from a directory that belongs to a python project package.
        And a source code with an object that needs an import statement.
        And that object belongs to the root of the python package.
    When: Fix code is run.
    Then: The import is done
    """
    source = dedent(
        """\
        fix_code()"""
    )
    fixed_source = dedent(
        """\
        from autoimport import fix_code


        fix_code()"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_autoimports_objects_defined_in___all__special_variable() -> None:
    """
    Given: Some missing packages in the __all__ variable
    When: Fix code is run.
    Then: The import is done
    """
    source = dedent(
        """\
        __all__ = ['fix_code']"""
    )
    fixed_source = dedent(
        """\
        from autoimport import fix_code


        __all__ = ['fix_code']"""
    )

    result = fix_code(source)

    assert result == fixed_source


@pytest.mark.parametrize(
    "source",
    [
        dedent(
            """\
            import os
            from typing import TYPE_CHECKING

            if TYPE_CHECKING:
                from .model import Book


            os.getcwd()


            def read_book(book: Book):
                pass"""
        ),
        dedent(
            """\
            from typing import TYPE_CHECKING

            if TYPE_CHECKING:
                foo = "bar"
            """
        ),
    ],
)
def test_fix_respects_type_checking_import_statements(source: str) -> None:
    """
    Given: Code with if TYPE_CHECKING imports
    When: Fix code is run.
    Then: The imports are not moved above the if statement.

    Related to https://github.com/lyz-code/autoimport/issues/231
    """
    result = fix_code(source)

    assert result == source


def test_fix_respects_multiparagraph_type_checking_import_statements() -> None:
    """
    Given: Code with two paragraphs of imports inside an if TYPE_CHECKING block
    When: Fix code is run.
    Then: The imports are not moved above the if statement.
    """
    source = dedent(
        """\
        import os
        from typing import TYPE_CHECKING

        if TYPE_CHECKING:
            from .model import Book

            from other import Other


        os.getcwd()


        def read_book(book: Book, other: Other):
            pass"""
    )

    result = fix_code(source)

    assert result == source


def test_fix_creates_the_typing_import() -> None:
    """
    Given: Code with no TYPE_CHECKING import statement
    When: Fix code is run.
    Then: The import is created.

    Related to https://github.com/lyz-code/autoimport/issues/231
    """
    source = dedent(
        """\
        if TYPE_CHECKING:
            foo = 'bar'"""
    )
    fixed_source = dedent(
        """\
        from typing import TYPE_CHECKING

        if TYPE_CHECKING:
            foo = 'bar'"""
    )

    result = fix_code(source)

    assert result == fixed_source


def test_fix_respects_try_except_in_import_statements() -> None:
    """
    Given: Code with try except statements in the imports.
    When: Fix code is run
    Then: The try except statements are respected
    """
    source = dedent(
        """\
        import os

        try:
            from typing import TypedDict  # noqa
        except ImportError:
            from mypy_extensions import TypedDict  # <=3.7


        os.getcwd()
        Movie = TypedDict('Movie', {'name': str, 'year': int})"""
    )

    result = fix_code(source)

    assert result == source


def test_fix_respects_leading_comments() -> None:
    """
    Given: Code with initial comments like shebang and editor configuration.
    When: Fix code is run
    Then: The comment statements are respected
    """
    source = dedent(
        '''\
        #!/usr/bin/env python3
        # -*- coding: latin-1 -*-
        """docstring"""
        print(os.path.exists("."))'''
    )
    desired_source = dedent(
        '''\
        #!/usr/bin/env python3
        # -*- coding: latin-1 -*-
        """docstring"""

        import os


        print(os.path.exists("."))'''
    )

    result = fix_code(source)

    assert result == desired_source


def test_fix_respects_leading_comments_with_new_lines() -> None:
    """
    Given: Code with initial comments with new lines and a trailing newline.
    When: Fix code is run.
    Then: The comment statements and trailing newline are respected.
    """
    source = dedent(
        '''\
        #!/usr/bin/env python3
        # -*- coding: latin-1 -*-

        # pylint: disable=foobar

        """

        This is the docstring.

        """

        import sys

        print(os.path.exists(sys.argv[1]))
        '''
    )
    desired_source = dedent(
        '''\
        #!/usr/bin/env python3
        # -*- coding: latin-1 -*-

        # pylint: disable=foobar

        """

        This is the docstring.

        """

        import sys

        import os


        print(os.path.exists(sys.argv[1]))
        '''
    )

    result = fix_code(source)

    assert result == desired_source


def test_fix_imports_dependency_only_once() -> None:
    """
    Given: Code with a line that uses a package three times.
    When: Fix code is run.
    Then: The dependency is imported only once
    """
    source = dedent(
        """\
        def f(x):
            return os.getcwd() + os.getcwd() + os.getcwd()
        """
    )
    desired_source = dedent(
        """\
        import os


        def f(x):
            return os.getcwd() + os.getcwd() + os.getcwd()
        """
    )

    result = fix_code(source)

    assert result == desired_source


def test_fix_doesnt_fail_on_empty_file() -> None:
    """
    Given: An empty file
    When: Fix code is run.
    Then: The output doesn't change
    """
    source = ""

    result = fix_code(source)

    assert result == source


def test_fix_not_remove_unused_imports() -> None:
    """
    Given: Code with imports, few being used, others not being used.
    When: Fix code is run.
    Then: Missing imports added, unused imports not removed.
    """
    source = dedent(
        """\
        import gzip
        import hashlib

        csv_writer = csv.DictWriter(filename, fieldnames=["name", "age"])
        gzip.open(filename, 'wb')
        """
    )
    desired_source = dedent(
        """\
        import gzip
        import hashlib

        import csv


        csv_writer = csv.DictWriter(filename, fieldnames=["name", "age"])
        gzip.open(filename, 'wb')
        """
    )

    result = fix_code(source, keep_unused_imports=True)

    assert result == desired_source


def test_file_that_only_has_unused_imports() -> None:
    """
    Given: A file that only has unused imports.
    When: Fix code is run.
    Then: The output should be a single empty line.
    """
    source = dedent(
        """\
        import os
        import sys
        """
    )

    result = fix_code(source)

    assert result == "\n"


def test_file_with_common_statement() -> None:
    """
    Given: Code with a commonly-used object.
    When: Fix code is run.
    Then: The appropriate import statement from the common_statements dict is added.
    """
    source = dedent(
        """\
        BeautifulSoup
        """
    )
    desired_source = dedent(
        """\
        from bs4 import BeautifulSoup


        BeautifulSoup
        """
    )

    result = fix_code(source)

    assert result == desired_source


def test_file_with_custom_common_statement() -> None:
    """
    Given: Code that uses an undefined object called `FooBar`.
    When:
        Fix code is run and a `config` dict is passed specifying `FooBar` as a common
        statement.
    Then:
        The appropriate import statement from the common_statements dict is added.
    """
    source = dedent(
        """\
        FooBar
        """
    )
    custom_config = {"common_statements": {"FooBar": "from baz_qux import FooBar"}}
    desired_source = dedent(
        """\
        from baz_qux import FooBar


        FooBar
        """
    )

    result = fix_code(source, config=custom_config)

    assert result == desired_source


def test_file_with_comment_in_import() -> None:
    """
    Given: Code with a comment on two import statements
    When: Fix code is run.
    Then: The unused import line is removed with it's comment
    """
    source = dedent(
        """\
        import os  # comment 1
        import sys  # comment 2

        os.getcwd()
        """
    )
    desired_source = dedent(
        """\
        import os  # comment 1


        os.getcwd()
        """
    )

    result = fix_code(source)

    assert result == desired_source


def test_file_with_comment_in_from_import() -> None:
    """
    Given: Code with a comment on two import statements
    When: Fix code is run.
    Then: The unused import line is removed with it's comment
    """
    source = dedent(
        """\
        import os  # comment 1
        from textwrap import dedent # comment 2

        os.getcwd()
        """
    )
    desired_source = dedent(
        """\
        import os  # comment 1


        os.getcwd()
        """
    )

    result = fix_code(source)

    assert result == desired_source


def test_file_with_comment_in_from_import_partial_remove() -> None:
    """
    Given: Code with a comment on an from import statement
    When: Fix code is run.
    Then: The unused dependency is removed but the comment is respected
    """
    source = dedent(
        """\
        from os import getcwd, chmod  # noqa: E0611

        getcwd()
        """
    )
    desired_source = dedent(
        """\
        from os import getcwd  # noqa: E0611


        getcwd()
        """
    )

    result = fix_code(source)

    assert result == desired_source


def test_file_with_comment_in_from_import_that_will_dissapear() -> None:
    """
    Given: Code with a comment on an from import statement that is to be deleted
    When: Fix code is run.
    Then: Everything is deleted
    """
    source = dedent(
        """\
        from os import getcwd, chmod  # noqa: E0611

        a = 1
        """
    )
    desired_source = dedent(
        """\
        a = 1
        """
    )

    result = fix_code(source)

    assert result == desired_source


def test_file_with_import_as() -> None:
    """
    Given: Code with an from x import y as z import statement
    When: Fix code is run.
    Then: The unused import line is removed
    """
    source = dedent(
        """\
        from subprocess import run as run
        """
    )

    result = fix_code(source)

    assert result == "\n"


def test_file_with_non_used_multiline_import() -> None:
    """
    Given: Code with a multiline from import where no one is used.
    When: Fix code is run.
    Then: The unused import line is removed
    """
    source = dedent(
        """\
        from foo import (
            bar,
            baz,
        )
        """
    )

    result = fix_code(source)

    assert result == "\n"


def test_file_with_import_and_seperator() -> None:
    """Ensure import lines with seperators are fixed correctly."""
    source = dedent(
        """
        a = 1
        import pdb;pdb.set_trace()
        b = 2
        """
    )
    expected = dedent(
        """
        import pdb


        a = 1
        pdb.set_trace()
        b = 2
        """
    ).replace("\n", "", 1)

    result = fix_code(source)

    assert result == expected


def test_file_with_import_and_seperator_indentation() -> None:
    """Ensure import lines with seperators are fixed correctly when indented."""
    source = dedent(
        """
        Class Person:
            import pdb; pdb.set_trace()
            def say_hi(self):
                print('hi')
        """
    )
    expected = dedent(
        """
        import pdb


        Class Person:
            pdb.set_trace()
            def say_hi(self):
                print('hi')
        """
    ).replace("\n", "", 1)

    result = fix_code(source)

    assert result == expected


def test_import_module_with_dot() -> None:
    """
    Given: An import file with an import with a dot
    When: running autoimport on the file
    Then: ValueError exception is not raised

    Tests https://github.com/lyz-code/autoimport/issues/225
    """
    source = dedent(
        """
        import my_module.m
        """
    )

    result = fix_code(source)

    assert result == "\n"


def test_respect_new_lines_between_imports_and_code() -> None:
    r"""
    Given: A file with two \n between imports and the code
    When: running autoimport on the file
    Then: the file is untouched

    For more info check https://github.com/lyz-code/autoimport/issues/219
    """
    source = dedent(
        """\
        import random


        def foo():
            print(random.random())
        """
    )

    result = fix_code(source)

    assert result == source



"""Tests for all entrypints modules."""

import re
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Sequence

import click
import pytest

from autoimport.entrypoints.cli import FileOrDir, flatten, get_files


@pytest.mark.parametrize(
    ("sequence", "expected"),
    [
        ((1, (2, 3, 4), 5), (1, 2, 3, 4, 5)),
        ([1, 2, 3, [4, 5, 6]], (1, 2, 3, 4, 5, 6)),
        ([["a", "b", "c"], "d"], ("a", "b", "c", "d")),
    ],
)
def test_flatten(sequence: Sequence[Any], expected: Sequence[Any]) -> None:
    """Test the flatten function works."""
    result = flatten(sequence)

    assert result == expected


def test_custom_param_type_works_with_dir(test_dir: Path) -> None:
    """Ensure the custom param type can be parsed a directory."""
    param_type = FileOrDir()

    result = param_type.convert(test_dir, None, None)

    for file_ in result:
        assert isinstance(file_, TextIOWrapper)
        assert re.match(r".*file[1-2].py", file_.name)
        file_.close()


def test_custom_param_type_works_with_file(test_dir: Path) -> None:
    """Ensure the custom param type can be parsed a file."""
    param_type = FileOrDir()

    result = param_type.convert(test_dir / "test_file1.py", None, None)

    assert re.match(r".*file[1-2].py", result[0].name)
    result[0].close()


@pytest.mark.parametrize("filename", ["h.py", "new_dir"])
def test_custom_param_type_with_non_existing_files(
    test_dir: Path, filename: str
) -> None:
    """Ensure an error occurs when a non existing file or dir is parsed."""
    param_type = FileOrDir()

    with pytest.raises(click.BadParameter) as error:
        param_type.convert(test_dir / filename, None, None)

    assert f"{filename}' does not exist" in error.value.args[0]


def test_get_files(test_dir: Path) -> None:
    """Ensure we can get all files recursively from a given directory."""
    result = get_files(str(test_dir))

    assert all(re.match(r".*file[1-2].py", file.name) for file in result)
    for file_path in result:
        file_path.close()



"""Test the extraction of package objects."""

from autoimport.model import extract_package_objects


def test_extraction_returns_the_package_functions() -> None:
    """
    Given: A package with functions.
    When: extract package objects is called
    Then: All the functions are extracted
    """
    result = extract_package_objects("autoimport")

    desired_objects = {
        "fix_code": "from autoimport import fix_code",
        "fix_files": "from autoimport import fix_files",
        "extract_package_objects": (
            "from autoimport.model import extract_package_objects"
        ),
    }
    for object_name, object_import_string in desired_objects.items():
        assert result[object_name] == object_import_string


def test_extraction_returns_the_package_classes() -> None:
    """
    Given: A package with classes.
    When: extract package objects is called.
    Then: All the classes are extracted.
    """
    result = extract_package_objects("autoimport")

    desired_objects = {
        "SourceCode": "from autoimport.model import SourceCode",
    }
    for object_name, object_import_string in desired_objects.items():
        assert result[object_name] == object_import_string


def test_extraction_returns_the_package_dictionaries() -> None:
    """
    Given: A package with dictionaries.
    When: extract package objects is called.
    Then: All the dictionaries are extracted.
    """
    result = extract_package_objects("autoimport")

    desired_objects = {
        "common_statements": "from autoimport.model import common_statements",
    }
    for object_name, object_import_string in desired_objects.items():
        assert result[object_name] == object_import_string


def test_extraction_returns_empty_dict_if_package_is_not_importable() -> None:
    """
    Given: Autoimport can't import the package.
    When: the extract package objects is called.
    Then: An empty directory is returned
    """
    result = extract_package_objects("inexistent")

    assert not result



"""Define all the orchestration functionality required by the program to work.

Classes and functions that connect the different domain model objects with the adapters
and handlers to achieve the program's purpose.
"""

from typing import Any, Dict, Optional, Tuple

from _io import TextIOWrapper

from autoimport.model import SourceCode


def fix_files(
    files: Tuple[TextIOWrapper, ...],
    config: Optional[Dict[str, Any]] = None,
    keep_unused_imports: bool = False,
) -> Optional[str]:
    """Fix the python source code of a list of files.

    If the input is taken from stdin, it will output the value to stdout.

    Args:
        files: List of files to fix.

    Returns:
        Fixed code retrieved from stdin or None.
    """
    for file_wrapper in files:
        source = file_wrapper.read()
        fixed_source = fix_code(source, config, keep_unused_imports)

        if fixed_source == source and file_wrapper.name != "<stdin>":
            continue

        try:
            # Click testing runner doesn't simulate correctly the reading from stdin
            # instead of setting the name attribute to `<stdin>` it gives an
            # AttributeError. But when you use it outside testing, no AttributeError
            # is raised and name has the value <stdin>. So there is no way of testing
            # this behaviour.
            if file_wrapper.name == "<stdin>":  # pragma no cover
                output = "output"
            else:
                output = "file"
        except AttributeError:
            output = "output"

        if output == "file":
            file_wrapper.seek(0)
            file_wrapper.write(fixed_source)
            file_wrapper.truncate()
            file_wrapper.close()
        else:
            return fixed_source

    return None


def fix_code(
    original_source_code: str,
    config: Optional[Dict[str, Any]] = None,
    keep_unused_imports: bool = False,
) -> str:
    """Fix python source code to correct import statements.

    It corrects these errors:

        * Add missed import statements.
        * Remove unused import statements.
        * Move import statements to the top.

    Args:
        original_source_code: Source code to be corrected.
        keep_unused_imports: If true, unused imports are retained.

    Returns:
        Corrected source code.
    """
    return SourceCode(
        original_source_code, config=config, keep_unused_imports=keep_unused_imports
    ).fix()
